from ovos_config.config import Configuration

from ovos_backend_client.backends.base import BackendType
from ovos_backend_client.backends.offline import OfflineBackend
from ovos_backend_client.backends.personal import PersonalBackend

API_REGISTRY = {
    BackendType.OFFLINE: {
        "admin": True,  # updates mycroft.conf if used
        "database": True,  # manages local files only
        "device": True,  # manages local files only
        "skill_settings": True,  # manages local files only
        "dataset": True,  # manages local files only
        "metrics": True,  # manages local files only
        "wolfram": True,  # key needs to be set
        "geolocate": True,  # nominatim - no key needed
        "stt": True,  # uses OPM and reads from mycroft.conf
        "owm": True,  # key needs to be set
        "email": True,  # smtp config needs to be set
        "oauth": True  # oauth PHAL plugin to register apps
    },
    BackendType.PERSONAL: {
        "admin": True,
        "database": True,  # requires ovos-personal-backend>=0.2.0
        "device": True,
        "skill_settings": True,
        "dataset": True,
        "metrics": True,
        "wolfram": True,
        "geolocate": True,
        "stt": True,
        "owm": True,
        "email": True,
        "oauth": True  # can use personal-backend-manager to register apps
    }
}


def get_backend_type(conf=None):
    conf = conf or Configuration()
    if "server" in conf:
        conf = conf["server"]
    if conf.get("disabled"):
        return BackendType.OFFLINE
    if "backend_type" in conf:
        return BackendType[conf["backend_type"].upper()]
    url = conf.get("url")
    if not url:
        return BackendType.OFFLINE
    return BackendType.PERSONAL


def get_backend_config(url=None, version="v1", identity_file=None, backend_type=None):
    config = Configuration()
    config_server = config.get("server") or {}
    if not url:
        url = config_server.get("url")
        version = config_server.get("version") or version
        backend_type = backend_type or get_backend_type(config)
    elif not backend_type:
        backend_type = get_backend_type({"url": url})

    if not url and backend_type:
        if backend_type == BackendType.PERSONAL:
            url = "http://0.0.0.0:6712"
        else:
            url = "http://127.0.0.1"

    return url, version, identity_file, backend_type



import enum
import json

from copy import deepcopy
from json_database import JsonStorageXDG, JsonDatabaseXDG
from ovos_config.config import Configuration
from ovos_config.locations import get_xdg_config_save_path, get_xdg_cache_save_path

from ovos_backend_client.identity import IdentityManager


class AudioTag(str, enum.Enum):
    UNTAGGED = "untagged"
    WAKE_WORD = "wake_word"
    SPEECH = "speech"
    NOISE = "noise"
    SILENCE = "silence"


class SpeakerTag(str, enum.Enum):
    UNTAGGED = "untagged"
    MALE = "male"
    FEMALE = "female"
    CHILDREN = "children"


class DatabaseModel:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def serialize(self):
        return self.__dict__

    @classmethod
    def deserialize(cls, kwargs):
        return cls(**kwargs)


class MetricModel(DatabaseModel):
    def __init__(self, metric_id, metric_type, meta=None, uuid="AnonDevice"):
        if isinstance(meta, str):
            meta = json.loads(meta)
        super().__init__(metric_id=metric_id, metric_type=metric_type,
                         meta=meta, uuid=uuid)


class WakeWordRecordingModel(DatabaseModel):
    def __init__(self, wakeword_id, transcription, path, meta=None,
                 uuid="AnonDevice", tag=AudioTag.UNTAGGED, speaker_type=SpeakerTag.UNTAGGED):
        if isinstance(meta, str):
            meta = json.loads(meta)
        super().__init__(wakeword_id=wakeword_id, transcription=transcription,
                         path=path, meta=meta or [], uuid=uuid,
                         tag=tag, speaker_type=speaker_type)


class UtteranceRecordingModel(DatabaseModel):
    def __init__(self, utterance_id, transcription, path, uuid="AnonDevice"):
        super().__init__(utterance_id=utterance_id, transcription=transcription, path=path, uuid=uuid)


class SkillSettingsModel(DatabaseModel):
    """ represents skill settings for a individual skill"""

    def __init__(self, skill_id, skill_settings=None,
                 meta=None, display_name=None, remote_id=None):
        remote_id = remote_id or skill_id
        if not remote_id.startswith("@"):
            remote_id = f"@|{remote_id}"
        if isinstance(meta, str):
            meta = json.loads(meta)
        super().__init__(skill_id=skill_id, skill_settings=skill_settings or {},
                         meta=meta or {}, display_name=display_name or skill_id, remote_id=remote_id)

    def store(self):
        with open(f"{get_xdg_config_save_path()}/skills/{self.skill_id}/settings.json" "w") as f:
            json.dump(self.skill_settings, f, indent=4, ensure_ascii=False)
        # TODO - autogen meta if needed (?)
        with open(f"{get_xdg_config_save_path()}/skills/{self.skill_id}/settingsmeta.json" "w") as f:
            json.dump(self.meta, f, indent=4, ensure_ascii=False)

    def serialize(self):
        # settings meta with updated placeholder values from settings
        # old style selene db stored skill settings this way
        meta = deepcopy(self.meta)
        for idx, section in enumerate(meta.get('sections', [])):
            for idx2, field in enumerate(section["fields"]):
                if "value" not in field:
                    continue
                if field["name"] in self.skill_settings:
                    meta['sections'][idx]["fields"][idx2]["value"] = self.skill_settings[field["name"]]
        return {'skillMetadata': meta,
                "skill_gid": self.remote_id,
                "display_name": self.display_name}

    @staticmethod
    def deserialize(data):
        if isinstance(data, str):
            data = json.loads(data)

        skill_json = {}
        skill_meta = data.get("skillMetadata") or {}
        for s in skill_meta.get("sections", []):
            for f in s.get("fields", []):
                if "name" in f and "value" in f:
                    val = f["value"]
                    if isinstance(val, str):
                        t = f.get("type", "")
                        if t == "checkbox":
                            if val.lower() == "true" or val == "1":
                                val = True
                            else:
                                val = False
                        elif t == "number":
                            if val == "False":
                                val = 0
                            elif val == "True":
                                val = 1
                            else:
                                val = float(val)
                        elif val.lower() in ["none", "null", "nan"]:
                            val = None
                        elif val == "[]":
                            val = []
                        elif val == "{}":
                            val = {}
                    skill_json[f["name"]] = val

        remote_id = data.get("skill_gid") or data.get("identifier")
        # this is a mess, possible keys seen by logging data
        # - @|XXX
        # - @{uuid}|XXX
        # - XXX

        # where XXX has been observed to be
        # - {skill_id}  <- ovos-core
        # - {msm_name} <- mycroft-core
        #   - {mycroft_marketplace_name} <- all default skills
        #   - {MycroftSkill.name} <- sometimes sent to msm (very uncommon)
        #   - {skill_id.split(".")[0]} <- fallback msm name
        # - XXX|{branch} <- append by msm (?)
        # - {whatever we feel like uploading} <- SeleneCloud utils
        fields = remote_id.split("|")
        skill_id = fields[0]
        if len(fields) > 1 and fields[0].startswith("@"):
            skill_id = fields[1]

        display_name = data.get("display_name") or \
                       skill_id.split(".")[0].replace("-", " ").replace("_", " ").title()

        return SkillSettingsModel(skill_id, skill_json, skill_meta, display_name,
                                  remote_id=remote_id)


class DeviceModel(DatabaseModel):
    """ global device settings
    represent some fields from mycroft.conf but also contain some extra fields
    """

    def __init__(self):
        identity = IdentityManager.get()

        default_ww = Configuration().get("listener", {}).get("wake_word", "hey_mycroft")
        default_tts = Configuration().get("tts", {}).get("module", "ovos-tts-plugin-mimic3-server")
        mail_cfg = Configuration().get("email", {})

        uuid = identity.uuid
        super().__init__(uuid=uuid, token=identity.access,
                         isolated_skills=True,
                         name=f"Device-{uuid}",
                         device_location="somewhere",  # indoor location
                         email=mail_cfg.get("recipient") or \
                               mail_cfg.get("smtp", {}).get("username"),
                         date_format=Configuration().get("date_format") or "DMY",
                         system_unit=Configuration().get("system_unit") or "metric",
                         time_format=Configuration().get("time_format") or "full",
                         opt_in=Configuration().get("opt_in") or False,
                         lang=Configuration().get("lang") or "en-us",
                         location=Configuration().get("location", {}),
                         default_tts=default_tts,
                         default_tts_cfg=Configuration().get("tts", {}).get(default_tts, {}),
                         default_ww=default_ww.replace(" ", "_"),
                         default_ww_cfg=Configuration().get("hotwords", {}).get(default_ww, {})
                         )

    @property
    def selene_device(self):
        return {
            "description": self.device_location,
            "uuid": self.uuid,
            "name": self.name,

            # not tracked / meaningless
            # just for api compliance with selene
            'coreVersion': "unknown",
            'platform': 'unknown',
            'enclosureVersion': "",
            "user": {"uuid": self.uuid}  # users not tracked
        }

    @property
    def selene_settings(self):
        # this endpoint corresponds to a mycroft.conf
        # location is usually grabbed in a separate endpoint
        # in here we return it in case downstream is
        # aware of this and wants to save 1 http call

        # NOTE - selene returns the full listener config
        # this SHOULD NOT be done, since backend has no clue of hardware downstream
        # we return only wake word config
        if self.default_ww and self.default_ww_cfg:
            ww_cfg = {self.default_ww: self.default_ww_cfg}
            listener = {"wakeWord": self.default_ww.replace(" ", "_")}
        else:
            ww_cfg = {}
            listener = {}

        tts_config = dict(self.default_tts_cfg)
        if "module" in tts_config:
            tts = tts_config.pop("module")
            tts_settings = {"module": tts, tts: tts_config}
        else:
            tts_settings = {}
        return {
            "dateFormat": self.date_format,
            "optIn": self.opt_in,
            "systemUnit": self.system_unit,
            "timeFormat": self.time_format,
            "uuid": self.uuid,
            "lang": self.lang,
            "location": self.location,
            "listenerSetting": listener,
            "hotwordsSetting": ww_cfg,  # not present in selene, parsed correctly by core
            'ttsSettings': tts_settings
        }


class JsonMetricDatabase(JsonDatabaseXDG):
    def __init__(self):
        super().__init__("ovos_metrics", xdg_folder=get_xdg_cache_save_path())

    def add_metric(self, metric_type=None, meta=None, uuid="AnonDevice"):
        metric_id = self.total_metrics() + 1
        metric = MetricModel(metric_id, metric_type, meta, uuid)
        self.add_item(metric)
        return metric

    def total_metrics(self):
        return len(self)

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            print(e)


class JsonWakeWordDatabase(JsonDatabaseXDG):
    def __init__(self):
        super().__init__("ovos_wakewords", xdg_folder=get_xdg_cache_save_path())

    def add_wakeword(self, transcription, path, meta=None,
                     uuid="AnonDevice", tag=AudioTag.UNTAGGED,
                     speaker_type=SpeakerTag.UNTAGGED):
        wakeword_id = self.total_wakewords()
        wakeword = WakeWordRecordingModel(wakeword_id,
                                          transcription,
                                          path, meta, uuid,
                                          tag, speaker_type)
        self.add_item(wakeword)
        return wakeword

    def get_wakeword(self, rec_id):
        try:
            ww = self[rec_id]
        except IndexError:
            ww = None
        if ww:
            return WakeWordRecordingModel.deserialize(ww)
        return None

    def update_wakeword(self, rec_id, transcription=None, path=None,
                        meta=None, tag=AudioTag.UNTAGGED,
                        speaker_type=SpeakerTag.UNTAGGED):
        ww = self.get_wakeword(rec_id)
        if not ww:
            return None
        if transcription:
            ww.transcription = transcription
        if path:
            ww.path = path
        if tag:
            ww.tag = tag
        if speaker_type:
            ww.speaker_type = speaker_type
        self[rec_id] = ww.serialize()
        return ww

    def delete_wakeword(self, rec_id):
        if self.get(rec_id):
            self.pop(rec_id)
            return True
        return False

    def total_wakewords(self):
        return len(self)

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            print(e)


class JsonUtteranceDatabase(JsonDatabaseXDG):
    def __init__(self):
        super().__init__("ovos_utterances", xdg_folder=get_xdg_cache_save_path())

    def add_utterance(self, transcription, path, uuid="AnonDevice"):
        utterance_id = self.total_utterances()
        utterance = UtteranceRecordingModel(utterance_id, transcription,
                                            path, uuid)
        self.add_item(utterance)

    def get_utterance(self, rec_id):
        utt = self[rec_id]
        if utt:
            return UtteranceRecordingModel.deserialize(utt)
        return None

    def update_utterance(self, rec_id, transcription):
        utt = self.get_utterance(rec_id)
        if not utt:
            return None
        utt.transcription = transcription
        self[rec_id] = utt.serialize()
        return utt

    def delete_utterance(self, rec_id):
        if self.get(rec_id):
            self.pop(rec_id)
            return True
        return False

    def total_utterances(self):
        return len(self)

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            print(e)


class OAuthTokenDatabase(JsonStorageXDG):
    """ This helper class creates ovos-config-assistant/ovos-backend-manager compatible json databases
        This allows users to use oauth even when not using a backend"""

    def __init__(self):
        super().__init__("ovos_oauth", xdg_folder=get_xdg_cache_save_path())

    def add_token(self, token_id, token_data):
        self[token_id] = token_data

    def update_token(self, token_id, token_data):
        self.add_token(token_id, token_data)

    def get_token(self, token_id):
        return self.get(token_id)

    def delete_token(self, token_id):
        if token_id in self:
            self.pop(token_id)
            return True
        return False

    def total_tokens(self):
        return len(self)


class OAuthApplicationDatabase(JsonStorageXDG):
    """ This helper class creates ovos-config-assistant/ovos-backend-manager compatible json databases
        This allows users to use oauth even when not using a backend"""

    def __init__(self):
        super().__init__("ovos_oauth_apps", xdg_folder=get_xdg_cache_save_path())

    def add_application(self, oauth_service,
                        client_id, client_secret,
                        auth_endpoint, token_endpoint, callback_endpoint, scope,
                        shell_integration=True):
        self[oauth_service] = {"oauth_service": oauth_service,
                               "client_id": client_id,
                               "client_secret": client_secret,
                               "auth_endpoint": auth_endpoint,
                               "token_endpoint": token_endpoint,
                               "callback_endpoint": callback_endpoint,
                               "scope": scope,
                               "shell_integration": shell_integration}

    def get_application(self, oauth_service):
        return self.get(oauth_service)

    def update_application(self, oauth_service,
                           client_id, client_secret,
                           auth_endpoint, token_endpoint,
                           callback_endpoint, scope, shell_integration=True):
        self.add_application(oauth_service,
                             client_id, client_secret,
                             auth_endpoint, token_endpoint,
                             callback_endpoint, scope, shell_integration)

    def delete_application(self, oauth_service):
        if oauth_service in self:
            self.pop(oauth_service)
            return True
        return False

    def total_apps(self):
        return len(self)



from ovos_backend_client.api import DeviceApi, WolframAlphaApi, OpenWeatherMapApi, STTApi, GeolocationApi
from ovos_backend_client.cloud import SeleneCloud, SecretSeleneCloud
from ovos_backend_client.config import RemoteConfigManager
from ovos_backend_client.pairing import is_paired, has_been_paired, check_remote_pairing, PairingManager
from ovos_backend_client.settings import RemoteSkillSettings



import abc
import json
from enum import Enum
from io import BytesIO, StringIO

import requests
from ovos_backend_client.database import SkillSettingsModel
from ovos_backend_client.identity import IdentityManager
from ovos_config.config import Configuration

try:
    from timezonefinder import TimezoneFinder
except ImportError:
    TimezoneFinder = None


class BackendType(str, Enum):
    OFFLINE = "offline"
    PERSONAL = "personal"


class AbstractBackend:

    def __init__(self, url, version="v1", identity_file=None, backend_type=BackendType.OFFLINE, credentials=None):
        self.backend_url = url
        self.backend_type = backend_type
        self._identity_file = identity_file
        self.backend_version = version
        if not url.startswith("http"):
            url = f"http://{url}"
        self.url = url
        self.credentials = credentials or {}

    @property
    def identity(self):
        if self._identity_file:
            # this is helpful if copying over the identity to a non-mycroft device
            # eg, selene call out proxy in local backend
            IdentityManager.set_identity_file(self._identity_file)
        return IdentityManager.get()

    @property
    def uuid(self):
        return self.identity.uuid

    @property
    def access_token(self):
        return self.identity.access

    @property
    def headers(self):
        return {"Device": self.uuid,
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.access_token}"}

    def check_token(self):
        if self.identity.is_expired():
            self.refresh_token()

    def refresh_token(self):
        pass

    def get(self, url=None, *args, **kwargs):
        url = url or self.url
        if not url.startswith("http"):
            url = f"http://{url}"
        headers = self.headers
        if "headers" in kwargs:
            headers.update(kwargs.pop("headers"))
        self.check_token()
        return requests.get(url, headers=headers, timeout=(3.05, 15), *args, **kwargs)

    def post(self, url=None, *args, **kwargs):
        url = url or self.url
        if not url.startswith("http"):
            url = f"http://{url}"
        headers = self.headers
        if "headers" in kwargs:
            headers.update(kwargs.pop("headers"))
        self.check_token()
        return requests.post(url, headers=headers, timeout=(3.05, 15), *args, **kwargs)

    def put(self, url=None, *args, **kwargs):
        url = url or self.url
        if not url.startswith("http"):
            url = f"http://{url}"
        headers = self.headers
        if "headers" in kwargs:
            headers.update(kwargs.pop("headers"))
        self.check_token()
        return requests.put(url, headers=headers, timeout=(3.05, 15), *args, **kwargs)

    def patch(self, url=None, *args, **kwargs):
        url = url or self.url
        if not url.startswith("http"):
            url = f"http://{url}"
        headers = self.headers
        if "headers" in kwargs:
            headers.update(kwargs.pop("headers"))
        self.check_token()
        return requests.patch(url, headers=headers, timeout=(3.05, 15), *args, **kwargs)

    def delete(self, url=None, *args, **kwargs):
        url = url or self.url
        if not url.startswith("http"):
            url = f"http://{url}"
        headers = self.headers
        if "headers" in kwargs:
            headers.update(kwargs.pop("headers"))
        self.check_token()
        return requests.delete(url, headers=headers, timeout=(3.05, 15), *args, **kwargs)

    # OWM Api
    @staticmethod
    def _get_lat_lon(**kwargs):
        lat = kwargs.get("latitude") or kwargs.get("lat")
        lon = kwargs.get("longitude") or kwargs.get("lon") or kwargs.get("lng")
        if not lat or not lon:
            cfg = Configuration().get("location", {}).get("coordinate", {})
            lat = cfg.get("latitude")
            lon = cfg.get("longitude")
        return lat, lon

    @staticmethod
    def owm_language(lang: str):
        """
        OWM supports 31 languages, see https://openweathermap.org/current#multi

        Convert Mycroft's language code to OpenWeatherMap's, if missing use english.

        Args:
            language_config: The Mycroft language code.
        """
        OPEN_WEATHER_MAP_LANGUAGES = (
            "af", "al", "ar", "bg", "ca", "cz", "da", "de", "el", "en", "es", "eu", "fa", "fi", "fr", "gl", "he", "hi",
            "hr", "hu", "id", "it", "ja", "kr", "la", "lt", "mk", "nl", "no", "pl", "pt", "pt_br", "ro", "ru",
            "se", "sk", "sl", "sp", "sr", "sv", "th", "tr", "ua", "uk", "vi", "zh_cn", "zh_tw", "zu"
        )
        special_cases = {"cs": "cz", "ko": "kr", "lv": "la"}
        lang_primary, lang_subtag = lang.split('-')
        if lang.replace('-', '_') in OPEN_WEATHER_MAP_LANGUAGES:
            return lang.replace('-', '_')
        if lang_primary in OPEN_WEATHER_MAP_LANGUAGES:
            return lang_primary
        if lang_subtag in OPEN_WEATHER_MAP_LANGUAGES:
            return lang_subtag
        if lang_primary in special_cases:
            return special_cases[lang_primary]
        return "en"

    @abc.abstractmethod
    def owm_get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def owm_get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def owm_get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def owm_get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        raise NotImplementedError()

    # Wolfram Alpha Api
    @abc.abstractmethod
    def wolfram_spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def wolfram_simple(self, query, units="metric", lat_lon=None, optional_params=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def wolfram_full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
        https://products.wolframalpha.com/api/documentation/
        Pods of interest
        - Input interpretation - Wolfram's determination of what is being asked about.
        - Name - primary name of
        """
        raise NotImplementedError()

    # Geolocation Api
    @staticmethod
    def _get_timezone(**kwargs):
        if TimezoneFinder:
            lat, lon = AbstractBackend._get_lat_lon(**kwargs)
            tz = TimezoneFinder().timezone_at(lng=float(lon), lat=float(lat))
            return {
                "name": tz.replace("/", " "),
                "code": tz
            }
        else:
            cfg = Configuration().get("location", {}).get("timezone")
            return cfg or {"name": "UTC", "code": "UTC"}

    @abc.abstractmethod
    def geolocation_get(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def reverse_geolocation_get(self, lat, lon):
        """Call the reverse geolocation endpoint.

        Args:
            lat (float): latitude
            lon (float): longitude

        Returns:
            str: JSON structure with lookup results
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def ip_geolocation_get(self, ip):
        """Call the geolocation endpoint.

        Args:
            ip (str): the ip address to lookup

        Returns:
            str: JSON structure with lookup results
        """
        raise NotImplementedError()

    # STT Api
    @abc.abstractmethod
    def stt_get(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum minutes to transcribe(?)

        Returns:
            dict: JSON structure with transcription results
        """
        raise NotImplementedError()

    # Device Api
    def device_get(self):
        """ Retrieve all device information from the web backend """
        return {"uuid": IdentityManager.get().uuid,
                "name": "AnonDevice",
                "description": "unknown",
                "coreVersion": "unknown",
                "enclosureVersion": "unknown",
                "platform": "ovos-backend-client",
                "user": {"uuid": "Anon"}
                }

    @abc.abstractmethod
    def device_get_settings(self):
        """ Retrieve device settings information from the web backend

        Returns:
            str: JSON string with user configuration information.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def device_get_code(self, state=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def device_activate(self, state, token,
                        core_version="unknown",
                        platform="unknown",
                        platform_build="unknown",
                        enclosure_version="unknown"):
        raise NotImplementedError()

    @abc.abstractmethod
    def device_update_version(self,
                              core_version="unknown",
                              platform="unknown",
                              platform_build="unknown",
                              enclosure_version="unknown"):
        raise NotImplementedError()

    @abc.abstractmethod
    def device_get_location(self):
        """ Retrieve device location information from the web backend

        Returns:
            str: JSON string with user location.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def device_upload_skills_data(self, data):
        """ Upload skills.json file. This file contains a manifest of installed
        and failed installations for use with the Marketplace.

        Args:
             data: dictionary with skills data from msm
        """
        raise NotImplementedError()

    # Metrics API
    @abc.abstractmethod
    def metrics_upload(self, name, data):
        """ upload metrics"""
        raise NotImplementedError()

    # OAuth API
    @abc.abstractmethod
    def oauth_refresh_token(self, dev_cred):
        """
            Refresh Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def oauth_get_token(self, dev_cred, auto_refresh=True):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier
                auto_refresh: refresh expired tokens automatically

            Returns:
                json string containing token and additional information
        """
        raise NotImplementedError()

    # Dataset API
    def dataset_upload_wake_word(self, audio, params, upload_url=None):
        """ upload wake word sample - url can be external to backend"""
        if not isinstance(audio, bytes):
            byte_data = audio.get_wav_data()
        else:
            byte_data = audio
        upload_url = upload_url or Configuration().get("listener", {}).get("wake_word_upload", {}).get("url")
        if upload_url:
            # upload to arbitrary server
            ww_files = {
                'audio': BytesIO(byte_data),
                'metadata': StringIO(json.dumps(params))
            }
            return self.post(upload_url, files=ww_files)
        return {}

    def dataset_upload_stt_recording(self, audio, params, upload_url=None):
        """ upload stt sample - url can be external to backend"""
        if not isinstance(audio, bytes):
            byte_data = audio.get_wav_data()
        else:
            byte_data = audio
        upload_url = upload_url or Configuration().get("listener", {}).get("utterance_upload", {}).get("url")
        if upload_url:
            # upload to arbitrary server
            ww_files = {
                'audio': BytesIO(byte_data),
                'metadata': StringIO(json.dumps(params))
            }
            return self.post(upload_url, files=ww_files)
        return {}

    # Email API
    @abc.abstractmethod
    def email_send(self, title, body, sender):
        raise NotImplementedError()

    # Skill settings api
    @abc.abstractmethod
    def skill_settings_upload(self, skill_settings):
        # list of SkillSettingsModel or dicts
        raise NotImplementedError()

    @abc.abstractmethod
    def skill_settings_download(self):
        # return list of SkillSettingsModel
        raise NotImplementedError()

    # Admin Api
    @abc.abstractmethod
    def admin_pair(self, uuid=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def admin_get_backend_config(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def admin_update_backend_config(self, config):
        raise NotImplementedError()

    @abc.abstractmethod
    def admin_set_device_location(self, uuid, loc):
        """
        loc = {
            "city": {
                "code": "Lawrence",
                "name": "Lawrence",
                "state": {
                    "code": "KS",
                    "name": "Kansas",
                    "country": {
                        "code": "US",
                        "name": "United States"
                    }
                }
            },
            "coordinate": {
                "latitude": 38.971669,
                "longitude": -95.23525
            },
            "timezone": {
                "code": "America/Chicago",
                "name": "Central Standard Time",
                "dstOffset": 3600000,
                "offset": -21600000
            }
        }
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def admin_set_device_prefs(self, uuid, prefs):
        """
        prefs = {"time_format": "full",
                "date_format": "DMY",
                "system_unit": "metric",
                "lang": "en-us",
                "wake_word": "hey_mycroft",
                "ww_config": {"phonemes": "HH EY . M AY K R AO F T",
                             "module": "ovos-ww-plugin-pocketsphinx",
                             "threshold": 1e-90},
                "tts_module": "ovos-tts-plugin-mimic",
                "tts_config": {"voice": "ap"}}
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def admin_set_device_info(self, uuid, info):
        """
        info = {"opt_in": True,
                "name": "my_device",
                "device_location": "kitchen",
                "email": "notifications@me.com",
                "isolated_skills": False,
                "lang": "en-us"}
        """
        raise NotImplementedError()

    # Database api
    @abc.abstractmethod
    def db_list_devices(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_device(self, uuid):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_device(self, uuid, name=None,
                         device_location=None, opt_in=False,
                         location=None, lang=None, date_format=None,
                         system_unit=None, time_format=None, email=None,
                         isolated_skills=False, ww_id=None, voice_id=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_device(self, uuid):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_device(self, uuid, token, name=None,
                       device_location="somewhere",
                       opt_in=Configuration().get("opt_in", False),
                       location=Configuration().get("location"),
                       lang=Configuration().get("lang"),
                       date_format=Configuration().get("date_format", "DMY"),
                       system_unit=Configuration().get("system_unit", "metric"),
                       time_format=Configuration().get("date_format", "full"),
                       email=None,
                       isolated_skills=False,
                       ww_id=None,
                       voice_id=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_shared_skill_settings(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_shared_skill_settings(self, skill_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_shared_skill_settings(self, skill_id,
                                        display_name=None,
                                        settings_json=None,
                                        metadata_json=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_shared_skill_settings(self, skill_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_shared_skill_settings(self, skill_id,
                                      display_name,
                                      settings_json,
                                      metadata_json):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_skill_settings(self, uuid):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_skill_settings(self, uuid, skill_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_skill_settings(self, uuid, skill_id,
                                 display_name=None,
                                 settings_json=None,
                                 metadata_json=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_skill_settings(self, uuid, skill_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_skill_settings(self, uuid, skill_id,
                               display_name,
                               settings_json,
                               metadata_json):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_oauth_apps(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_oauth_app(self, token_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_oauth_app(self, token_id, client_id=None, client_secret=None,
                            auth_endpoint=None, token_endpoint=None,
                            callback_endpoint=None, scope=None, shell_integration=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_oauth_app(self, token_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_oauth_app(self, token_id, client_id, client_secret,
                          auth_endpoint, token_endpoint,
                          callback_endpoint, scope, shell_integration=True):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_oauth_tokens(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_oauth_token(self, token_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_oauth_token(self, token_id, token_data):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_oauth_token(self, token_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_oauth_token(self, token_id, token_data):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_stt_recordings(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_stt_recording(self, rec_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_stt_recording(self, rec_id, transcription=None, metadata=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_stt_recording(self, rec_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_stt_recording(self, byte_data, transcription, metadata=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_ww_recordings(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_ww_recording(self, rec_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_ww_recording(self, rec_id, transcription=None, metadata=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_ww_recording(self, rec_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_ww_recording(self, byte_data, transcription, metadata=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_metrics(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_metric(self, metric_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_metric(self, metric_id, metadata):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_metric(self, metric_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_metric(self, metric_type, metadata):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_ww_definitions(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_ww_definition(self, ww_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_ww_definition(self, ww_id, name, lang, ww_config, plugin):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_ww_definition(self, ww_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_ww_definition(self, name, lang, ww_config, plugin):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_list_voice_definitions(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_get_voice_definition(self, voice_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_update_voice_definition(self, voice_id, name=None, lang=None, plugin=None,
                                   tts_config=None, offline=None, gender=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_delete_voice_definition(self, voice_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def db_post_voice_definition(self, name, lang, plugin,
                                 tts_config, offline, gender=None):
        raise NotImplementedError()

    # DEPRECATED APIS
    @property
    def is_subscriber(self):
        """
            status of subscription. True if device is connected to a paying
            subscriber.
        """
        ## DEPRECATED - compat only for old devices
        try:
            return self.device_get_subscription().get('@type') != 'free'
        except Exception:
            # If can't retrieve, assume not paired and not a subscriber yet
            return False

    def device_report_metric(self, name, data):
        ## DEPRECATED - compat only for old devices
        return self.metrics_upload(name, data)

    def device_get_oauth_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        ## DEPRECATED - compat only for old devices
        raise self.oauth_get_token(dev_cred)

    def device_get_skill_settings_v1(self):
        """ old style bidirectional skill settings api, still available!"""
        ## DEPRECATED - compat only for old devices
        return [s.serialize() for s in self.skill_settings_download()]

    def device_put_skill_settings_v1(self, data=None):
        """ old style bidirectional skill settings api, still available!"""
        ## DEPRECATED - compat only for old devices
        s = SkillSettingsModel.deserialize(data)
        settings = [s]
        self.skill_settings_upload(settings)
        return {}

    def device_get_skill_settings(self):
        """Get the remote skill settings for all skills on this device."""
        ## DEPRECATED - compat only for old devices
        return {s.skill_id: s.skill_settings for s in self.skill_settings_download()}

    def device_upload_skill_metadata(self, settings_meta):
        """Upload skill metadata.

        Args:
            settings_meta (dict): skill info and settings in JSON format
        """
        ## DEPRECATED - compat only for old devices
        s = SkillSettingsModel.deserialize(settings_meta)
        settings = self.device_get_skill_settings()
        old_s = settings.get(s.skill_id)
        if old_s:  # keep old settings value, update meta values only
            s.skill_settings = old_s.skill_settings
        self.device_put_skill_settings_v1(s.serialize())

    def device_send_email(self, title, body, sender):
        ## DEPRECATED - compat only for old devices
        return self.email_send(title, body, sender)

    def device_upload_wake_word_v1(self, audio, params, upload_url=None):
        """ upload precise wake word V1 endpoint - url can be external to backend"""
        return self.dataset_upload_wake_word(audio, params, upload_url)

    def device_upload_wake_word(self, audio, params):
        """ upload precise wake word V2 endpoint - integrated with device api"""
        return self.dataset_upload_wake_word(audio, params)

    def device_get_subscription(self):
        """
            Get information about type of subscription this unit is connected
            to.

            Returns: dictionary with subscription information
        """
        ## DEPRECATED - compat only for old devices
        return {"@type": "free"}

    def device_get_subscriber_voice_url(self, voice=None, arch=None):
        ## DEPRECATED - compat only for old devices
        return None



import json
import os
import re
from copy import deepcopy
from os import makedirs
from os.path import dirname, expanduser, join, isfile, isdir

from json_database import JsonStorage
from ovos_config import Configuration
from ovos_config.config import get_xdg_config_save_path
from ovos_config.locations import get_xdg_data_save_path, get_xdg_data_dirs
from ovos_utils import camel_case_split
from ovos_backend_client.database import SkillSettingsModel
import ovos_backend_client.api as _api


def get_display_name(skill_name: str):
    """Splits camelcase and removes leading/trailing "skill"."""
    skill_name = skill_name.replace("_", " ").replace("-", " ")
    skill_name = re.sub(r'(^[Ss]kill|[Ss]kill$)', '', skill_name)
    return camel_case_split(skill_name).title().strip()


def get_local_settings():
    settings_path = f"{get_xdg_config_save_path()}/skills"
    if not isdir(settings_path):
        return []
    settings = {}
    meta = {}
    all_settings = []
    for skill_id in os.listdir(settings_path):
        s = f"{settings_path}/{skill_id}/settings.json"
        if isfile(s):
            with open(s) as f:
                settings = json.load(f)
        s = f"{settings_path}/{skill_id}/settingsmeta.json"
        if isfile(s):
            meta = json.load(f)

        display_name = skill_id.split(".")[-1].replace("_", "").replace("-", "").title()
        s = SkillSettingsModel(skill_id=skill_id, meta=meta,
                               skill_settings=settings, display_name=display_name)
        all_settings.append(s)
    return all_settings


class RemoteSkillSettings:
    """ WARNING: selene backend does not use proper skill_id, if you have
    skills with same name but different author settings will overwrite each
    other on the backend, THIS CLASS IS NOT 100% SAFE in mycroft-core

    mycroft-core uses msm to generate weird metadata, removes author and munges github branch names into id
    see: https://github.com/MycroftAI/mycroft-skills-manager/blob/master/msm/skill_entry.py#L145

    ovos-core uses the proper deterministic skill_id and can be used safely
    if running in mycroft-core you want to use remote_id=self.settings_meta.skill_gid

    you can define arbitrary strings as skill_id to use this as a datastore

    skill matching is currently done by checking "if {skill_id} in string"
    """

    def __init__(self, skill_id, settings=None, meta=None, url=None, version="v1", remote_id=None):
        self.api = _api.DeviceApi(url, version)
        self.skill_id = skill_id
        self.identifier = remote_id or \
                          self.selene_gid if not skill_id.startswith("@") else skill_id
        self.settings = settings or {}
        self.meta = meta or {}
        self.local_path = join(get_xdg_config_save_path(), 'skills', self.skill_id, 'settings.json')
        if not self.settings:
            self.load()

    @property
    def selene_gid(self):
        if self.api.identity.uuid:
            return f'@{self.api.identity.uuid}|{self.skill_id}'
        return f'@|{self.skill_id}'

    @staticmethod
    def _settings2meta(settings):
        """ generates basic settingsmeta fields"""
        fields = []
        for k, v in settings.items():
            if k.startswith("_"):
                continue
            label = k.replace("-", " ").replace("_", " ").title()
            if isinstance(v, bool):
                fields.append({
                    "name": k,
                    "type": "checkbox",
                    "label": label,
                    "value": str(v).lower()
                })
            if isinstance(v, str):
                fields.append({
                    "name": k,
                    "type": "text",
                    "label": label,
                    "value": v
                })
            if isinstance(v, int) or isinstance(v, float):
                fields.append({
                    "name": k,
                    "type": "number",
                    "label": label,
                    "value": str(v)
                })
        return fields

    def generate_meta(self):
        """ auto generate settings meta info for any valid value defined in settings but missing in meta"""
        names = []
        if "sections" not in self.meta:
            self.meta["sections"] = []
        for s in self.meta["sections"]:
            names += [f["name"] for f in s.get("fields", [])]
        new_meta = self._settings2meta(
            {k: v for k, v in self.settings.items()
             if k not in names and not k.startswith("_")})
        for idx, s in enumerate(self.meta["sections"]):
            if s.get("name") == "Skill Settings":
                self.meta["sections"][idx] = {"name": "Skill Settings", "fields": new_meta}
                break
        else:
            self.meta["sections"].append({"name": "Skill Settings", "fields": new_meta})
        # TODO auto update in backend ?

    def download(self, filter_uuid=False):
        """
        download skill settings for this skill from selene

        WARNING: mycroft-core does not use proper skill_id, if you have
        skills with same name but different author settings will overwrite each
        other on the backend, THIS METHOD IS NOT SAFE in mycroft-core

        mycroft-core uses msm to generate weird metadata, removes author and munges github branch names into id
        if running in mycroft-core you want to use remote_id=self.settings_meta.skill_gid

        ovos-core uses the proper deterministic skill_id and can be used safely
        """
        data = self.api.get_skill_settings_v1()

        def match_settings(x, against):
            # this is a mess, possible keys seen by logging data
            # - @|XXX
            # - @{uuid}|XXX
            # - XXX

            # where XXX has been observed to be
            # - {skill_id}  <- ovos-core
            # - {msm_name} <- mycroft-core
            #   - {mycroft_marketplace_name} <- all default skills
            #   - {MycroftSkill.name} <- sometimes sent to msm (very uncommon)
            #   - {skill_id.split(".")[0]} <- fallback msm name
            # - XXX|{branch} <- append by msm (?)
            # - {whatever we feel like uploading} <- SeleneCloud utils

            for sets in x:
                fields = sets["identifier"].split("|")
                skill_id = fields[0]
                uuid = None
                if len(fields) >= 2 and fields[0].startswith("@"):
                    uuid = fields[0].replace("@", "")
                    skill_id = fields[1]

                # setting belong to another device
                if uuid and uuid != self.api.uuid:
                    # shared_settings flag
                    # do not return settings for same skill from other devices
                    # this should not be relied on, no assurance that uuid is part of gid
                    # selene assumes settings are shared, only cares about skill versions not devices
                    # TODO - proper integration with local backend (?)
                    if filter_uuid:
                        continue

                if skill_id == against or sets["identifier"] == against:
                    return self.deserialize(sets)

        s = match_settings(data, self.identifier) or \
            match_settings(data, self.skill_id)

        if s:
            self.meta = s.meta
            self.settings = s.skill_settings
            # update actual identifier from selene
            self.identifier = s.identifier

    def upload(self):
        data = self.serialize()
        return self.api.put_skill_settings_v1(data)

    def upload_meta(self):
        self.api.upload_skill_metadata(self.meta)

    def load(self):
        if not isfile(self.local_path):
            self.settings = {}
        else:
            with open(self.local_path) as f:
                self.settings = json.load(f)

    def store(self):
        makedirs(dirname(self.local_path), exist_ok=True)
        with open(self.local_path, "w") as f:
            json.dump(self.settings, f, indent=2)

    def get(self, key):
        return self.settings.get(key)

    def __str__(self):
        return str(self.settings)

    def __setitem__(self, key, value):
        self.settings[key] = value

    def __getitem__(self, item):
        return self.settings.get(item)

    def __dict__(self):
        return self.serialize()

    def serialize(self):
        meta = deepcopy(self.meta)
        for idx, section in enumerate(meta.get('sections', [])):
            for idx2, field in enumerate(section["fields"]):
                if "value" not in field:
                    continue
                if field["name"] in self.settings:
                    val = self.settings[field["name"]]
                    meta['sections'][idx]["fields"][idx2]["value"] = str(val)
        return {'skillMetadata': meta,
                "skill_gid": self.identifier,
                "display_name": get_display_name(self.skill_id)}

    def deserialize(self, data):
        if isinstance(data, str):
            data = json.loads(data)

        skill_json = {}
        skill_meta = data.get("skillMetadata") or {}
        for s in skill_meta.get("sections", []):
            for f in s.get("fields", []):
                if "name" in f and "value" in f:
                    val = f["value"]
                    if isinstance(val, str):
                        t = f.get("type", "")
                        if t == "checkbox":
                            if val.lower() == "true" or val == "1":
                                val = True
                            else:
                                val = False
                        elif t == "number":
                            val = float(val)
                        elif val.lower() in ["none", "null", "nan"]:
                            val = None
                        elif val == "[]":
                            val = []
                        elif val == "{}":
                            val = {}

                    skill_json[f["name"]] = val

        remote_id = data.get("skill_gid") or \
                    data.get("identifier")  # deprecated

        fields = remote_id.split("|")
        skill_id = fields[0]
        if len(fields) > 1 and fields[0].startswith("@"):
            skill_id = fields[1]
        return RemoteSkillSettings(skill_id, skill_json, skill_meta, remote_id=remote_id,
                                   url=self.api.backend_url, version=self.api.backend_version)

    def __enter__(self):
        self.load()
        self.download()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.generate_meta()
        self.upload()
        self.store()


class SeleneSkillsManifest(JsonStorage):
    """dict subclass with save/load support

    This dictionary contains the metadata expected by selene backend
    This data is used to populate entries in selene skill settings page

    """

    def __init__(self, api=None):
        path = os.path.join(get_xdg_data_save_path(), 'skills.json')
        super().__init__(path, disable_lock=True)
        if "skills" not in self:
            self["skills"] = []
            self.store()
        self.api = api or _api.DeviceApi()

    @staticmethod
    def _get_default_skills_directory(conf=None):
        """ return default directory to scan for skills

        users can define the data directory in mycroft.conf
        the skills folder name (relative to data_dir) can also be defined there

        NOTE: folder name also impacts all XDG skill directories!

        {
            "data_dir": "/opt/mycroft",
            "skills": {
                "directory": "skills"
            }
        }

        Args:
            conf (dict): mycroft.conf dict, will be loaded automatically if None
        """
        conf = conf or Configuration()
        data_dir = conf.get("data_dir") or "/opt/mycroft"
        folder = conf["skills"].get("directory") or "skills"

        skills_folder = f"{data_dir}/{folder}"

        # create folder if needed
        try:
            makedirs(skills_folder, exist_ok=True)
        except PermissionError:  # old style /opt/mycroft/skills not available
            skills_folder = f"{get_xdg_data_save_path()}/{folder}"
            makedirs(skills_folder, exist_ok=True)

        return expanduser(skills_folder)

    @staticmethod
    def _get_skill_directories(conf=None):
        """ returns list of skill directories ordered by expected loading order

        This corresponds to:
        - XDG_DATA_DIRS
        - default directory (see get_default_skills_directory method for details)
        - user defined extra directories

        Each directory contains individual skill folders to be loaded

        If a skill exists in more than one directory (same folder name) previous instances will be ignored
            ie. directories at the end of the list have priority over earlier directories

        NOTE: empty folders are interpreted as disabled skills

        new directories can be defined in mycroft.conf by specifying a full path
        each extra directory is expected to contain individual skill folders to be loaded

        the xdg folder name can also be changed, it defaults to "skills"
            eg. ~/.local/share/mycroft/FOLDER_NAME

        {
            "skills": {
                "directory": "skills",
                "extra_directories": ["path/to/extra/dir/to/scan/for/skills"]
            }
        }

        Args:
            conf (dict): mycroft.conf dict, will be loaded automatically if None
        """
        # the contents of each skills directory must be individual skill folders
        # we are still dependent on the mycroft-core structure of skill_id/__init__.py

        conf = conf or Configuration()
        folder = conf["skills"].get("directory") or "skills"

        # load all valid XDG paths
        skill_locations = list(reversed(
            [os.path.join(p, folder) for p in get_xdg_data_dirs()]
        ))

        # load the default skills folder
        default = SeleneSkillsManifest._get_default_skills_directory(conf)
        if default not in skill_locations:
            skill_locations.append(default)

        # load additional explicitly configured directories
        conf = conf.get("skills") or {}
        # extra_directories is a list of directories containing skill subdirectories
        # NOT a list of individual skill folders
        skill_locations += conf.get("extra_directories") or []
        return skill_locations

    def device_skill_state_hash(self):
        return hash(json.dumps(self, sort_keys=True))

    def add_skill(self, skill_id):
        if self.api.identity.uuid:
            skill_gid = f'@{self.api.identity.uuid}|{skill_id}'
        else:
            skill_gid = f'@|{skill_id}'
        skill = {
            "name": skill_id,
            "origin": "non-msm",
            "beta": True,
            "status": 'active',
            "installed": 0,
            "updated": 0,
            "installation": 'installed',
            "skill_gid": skill_gid
        }
        if "skills" not in self:
            self["skills"] = []
        self["skills"].append(skill)

    def get_skill_state(self, skill_id):
        """Find a skill entry in the device skill state and returns it."""
        for skill_state in self.get('skills', []):
            if skill_state.get('name') == skill_id:
                return skill_state
        return {}

    def scan_skills(self, skill_dirs=None):
        # TODO - this should go into ovos_utils
        skill_dirs = skill_dirs or self._get_skill_directories()
        for directory in skill_dirs:
            if not isdir(directory):
                continue
            for skill_id in os.listdir(directory):
                skill_init = join(directory, skill_id, "__init__.py")
                if isfile(skill_init):
                    self.add_skill(skill_id)
        # TODO - plugin skills
        self.store()


if __name__ == "__main__":
    s = RemoteSkillSettings("mycroft-date-time")
    print(s.api.get_skill_settings_v1())
    s.download()
    print(s)
    s.skill_settings["not"] = "yes"  # ignored, not in meta
    s.skill_settings["show_time"] = True
    s.upload()
    s.download()
    print(s)
    s.skill_settings["not"] = "yes"
    s.generate_meta()  # now in meta
    s.skill_settings["not"] = "no"
    s.skill_settings["show_time"] = False
    s.upload()
    s.download()
    print(s)



import json
import os
import time
from os import listdir, makedirs, remove
from os.path import isfile, join
from tempfile import NamedTemporaryFile
from uuid import uuid4

import requests
from oauthlib.oauth2 import WebApplicationClient
from ovos_backend_client.backends.base import AbstractBackend, BackendType
from ovos_backend_client.database import JsonMetricDatabase, JsonWakeWordDatabase, \
    SkillSettingsModel, OAuthTokenDatabase, OAuthApplicationDatabase, DeviceModel, JsonUtteranceDatabase
from ovos_backend_client.identity import IdentityManager
from ovos_backend_client.settings import get_local_settings
from ovos_config.config import Configuration, update_mycroft_config, get_xdg_config_save_path
from ovos_config.locations import USER_CONFIG, get_xdg_data_save_path, xdg_data_home
from ovos_utils import timed_lru_cache
from ovos_utils.log import LOG
from ovos_utils.network_utils import get_external_ip
from ovos_utils.smtp_utils import send_smtp

try:
    from ovos_plugin_manager.tts import get_voices, get_voice_id
    from ovos_plugin_manager.wakewords import get_ww_id, get_wws
except ImportError:
    from hashlib import md5


    def get_ww_id(plugin_name, ww_name, ww_config):
        ww_hash = md5(json.dumps(ww_config, sort_keys=True).encode("utf-8")).hexdigest()
        return f"{plugin_name}_{ww_name}_{ww_hash}"


    def get_voice_id(plugin_name, lang, tts_config):
        tts_hash = md5(json.dumps(tts_config, sort_keys=True).encode("utf-8")).hexdigest()
        return f"{plugin_name}_{lang}_{tts_hash}"


    def get_voices():
        return []


    def get_wws():
        return []


class OfflineBackend(AbstractBackend):

    def __init__(self, url="127.0.0.1", version="v1", identity_file=None, credentials=None):
        super().__init__(url, version, identity_file, BackendType.OFFLINE, credentials)
        self.stt = None

    # OWM API
    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def owm_get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/onecall"
        response = self.get(url, params=params)
        return response.json()

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def owm_get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/weather"
        response = self.get(url, params=params)
        return response.json()

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def owm_get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/forecast"
        response = self.get(url, params=params)
        return response.json()

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def owm_get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/forecast/daily"
        response = self.get(url, params=params)
        return response.json()

    # Wolfram Alpha Api
    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def wolfram_spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = 'https://api.wolframalpha.com/v1/spoken'
        params["appid"] = self.credentials["wolfram"]
        return self.get(url, params=params).text

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def wolfram_simple(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = 'https://api.wolframalpha.com/v1/simple'
        params["appid"] = self.credentials["wolfram"]
        return self.get(url, params=params).text

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def wolfram_full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
        https://products.wolframalpha.com/api/documentation/
        Pods of interest
        - Input interpretation - Wolfram's determination of what is being asked about.
        - Name - primary name of
        """
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "units": units,
                  "mode": "Default",
                  "format": "image,plaintext",
                  "geolocation": "{},{}".format(*lat_lon),
                  "output": "json",
                  **optional_params}
        url = 'https://api.wolframalpha.com/v2/query'
        params["appid"] = self.credentials["wolfram"]
        data = self.get(url, params=params)
        return data.json()

    # Geolocation Api
    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def geolocation_get(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        url = "https://nominatim.openstreetmap.org/search"
        data = self.get(url, params={"q": location, "format": "json", "limit": 1}).json()[0]
        lat = data.get("lat")
        lon = data.get("lon")

        if lat and lon:
            return self.reverse_geolocation_get(lat, lon)

        url = "https://nominatim.openstreetmap.org/details.php"
        details = self.get(url, params={"osmid": data['osm_id'], "osmtype": data['osm_type'][0].upper(),
                                        "format": "json"}).json()

        # if no addresstags are present for the location an empty list is sent instead of a dict
        tags = details.get("addresstags") or {}

        place_type = details["extratags"].get("linked_place") or details.get("category") or data.get(
            "type") or data.get("class")
        name = details["localname"] or details["names"].get("name") or details["names"].get("official_name") or data[
            "display_name"]
        cc = details["country_code"] or tags.get("country") or details["extratags"].get('ISO3166-1:alpha2') or ""
        # TODO - lang support, official name is reported in various langs
        location = {
            "address": data["display_name"],
            "city": {
                "code": tags.get("postcode") or
                        details["calculated_postcode"] or "",
                "name": name if place_type == "city" else "",
                "state": {
                    "code": tags.get("state_code") or
                            details["calculated_postcode"] or "",
                    "name": name if place_type == "state" else tags.get("state"),
                    "country": {
                        "code": cc.upper(),
                        "name": name if place_type == "country" else ""  # TODO - country code to name
                    }
                }
            },
            "coordinate": {
                "latitude": lat,
                "longitude": lon
            }
        }
        if "timezone" not in location:
            location["timezone"] = self._get_timezone(lon=lon, lat=lat)
        return location

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def reverse_geolocation_get(self, lat, lon):
        """Call the reverse geolocation endpoint.

        Args:
            lat (float): latitude
            lon (float): longitude

        Returns:
            str: JSON structure with lookup results
        """
        url = "https://nominatim.openstreetmap.org/reverse"
        details = self.get(url, params={"lat": lat, "lon": lon, "format": "json"}).json()
        address = details.get("address")
        location = {
            "address": details["display_name"],
            "city": {
                "code": address.get("postcode") or "",
                "name": address.get("city") or
                        address.get("village") or
                        address.get("town") or
                        address.get("hamlet") or
                        address.get("county") or "",
                "state": {
                    "code": address.get("state_code") or
                            address.get("ISO3166-2-lvl4") or
                            address.get("ISO3166-2-lvl6")
                            or "",
                    "name": address.get("state") or
                            address.get("county")
                            or "",
                    "country": {
                        "code": address.get("country_code", "").upper() or "",
                        "name": address.get("country") or "",
                    }
                }
            },
            "coordinate": {
                "latitude": details.get("lat") or lat,
                "longitude": details.get("lon") or lon
            }
        }
        if "timezone" not in location:
            location["timezone"] = self._get_timezone(
                lat=details.get("lat") or lat,
                lon=details.get("lon") or lon)
        return location

    @timed_lru_cache(seconds=600)  # cache results for 10 mins
    def ip_geolocation_get(self, ip):
        """Call the geolocation endpoint.

        Args:
            ip (str): the ip address to lookup

        Returns:
            str: JSON structure with lookup results
        """
        if not ip or ip in ["0.0.0.0", "127.0.0.1"]:
            ip = get_external_ip()
        fields = "status,country,countryCode,region,regionName,city,lat,lon,timezone,query"
        data = requests.get("http://ip-api.com/json/" + ip,
                            params={"fields": fields}).json()
        region_data = {"code": data["region"],
                       "name": data["regionName"],
                       "country": {
                           "code": data["countryCode"],
                           "name": data["country"]}}
        city_data = {"code": data["city"],
                     "name": data["city"],
                     "state": region_data}
        timezone_data = {"code": data["timezone"],
                         "name": data["timezone"]}
        coordinate_data = {"latitude": float(data["lat"]),
                           "longitude": float(data["lon"])}
        return {"city": city_data,
                "coordinate": coordinate_data,
                "timezone": timezone_data}

    # Device Api
    def device_get(self):
        """ Retrieve all device information from the json db"""
        device = DeviceModel()
        return device.selene_device

    def device_get_settings(self):
        """ Retrieve device settings information from the json db

        Returns:
            str: JSON string with user configuration information.
        """
        device = DeviceModel()
        return device.selene_settings

    def device_get_code(self, state=None):
        return "ABCDEF"  # dummy data

    def device_activate(self, state, token,
                        core_version="unknown",
                        platform="unknown",
                        platform_build="unknown",
                        enclosure_version="unknown"):
        identity = self.admin_pair(state)
        return identity

    def device_update_version(self,
                              core_version="unknown",
                              platform="unknown",
                              platform_build="unknown",
                              enclosure_version="unknown"):
        pass  # irrelevant info

    def device_get_location(self):
        """ Retrieve device location information from Configuration

        Returns:
            str: JSON string with user location.
        """
        return Configuration().get("location") or {}

    def device_upload_skills_data(self, data):
        """ Upload skills.json file. This file contains a manifest of installed
        and failed installations for use with the Marketplace.

        Args:
             data: dictionary with skills data from msm
        """
        pass

    # Metrics API
    def metrics_upload(self, name, data):
        """ upload metrics"""
        return self.db_post_metric(name, data)

    # Skill settings api
    def skill_settings_upload(self, skill_settings):
        # update on disk, settings already local
        for s in skill_settings:
            s.store()

    def skill_settings_download(self):
        # settings already local
        return get_local_settings()

    # Dataset API
    def dataset_upload_wake_word(self, audio, params, upload_url=None):
        """ upload wake word sample - url can be external to backend"""
        if not isinstance(audio, bytes):
            byte_data = audio.get_wav_data()
        else:
            byte_data = audio
        if Configuration().get("listener", {}).get('record_wake_words'):
            self.db_post_ww_recording(byte_data, params["name"], params)
        return super().dataset_upload_wake_word(audio, params, upload_url)

    def dataset_upload_stt_recording(self, audio, params, upload_url=None):
        """ upload stt sample - url can be external to backend"""
        if not isinstance(audio, bytes):
            byte_data = audio.get_wav_data()
        else:
            byte_data = audio
        if Configuration().get("listener", {}).get('record_utterances'):
            self.db_post_stt_recording(byte_data, params["transcription"], params)
        return super().dataset_upload_stt_recording(audio, params, upload_url)

    # Email API
    def email_send(self, title, body, sender):
        """ will raise KeyError if SMTP not configured in mycroft.conf"""
        body += f"\n\nsent by: {sender}"  # append skill_id info to body

        mail_config = self.credentials["email"]

        smtp_config = mail_config["smtp"]
        user = smtp_config["username"]
        pswd = smtp_config["password"]
        host = smtp_config["host"]
        port = smtp_config.get("port", 465)

        recipient = mail_config.get("recipient") or user

        send_smtp(user, pswd,
                  user, recipient,
                  title, body,
                  host, port)

    # OAuth API
    def oauth_refresh_token(self, dev_cred):
        """
            Refresh Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        # Load all needed data for refresh
        with OAuthApplicationDatabase() as db:
            app_data = db.get(dev_cred)
        with OAuthTokenDatabase() as db:
            token_data = db.get(dev_cred)

        if (app_data is None or
                token_data is None or 'refresh_token' not in token_data):
            LOG.warning("Token data doesn't contain a refresh token and "
                        "cannot be refreshed.")
            return

        refresh_token = token_data["refresh_token"]

        # Fall back to token endpoint if no specific refresh endpoint
        # has been set
        token_endpoint = app_data["token_endpoint"]

        client_id = app_data["client_id"]
        client_secret = app_data["client_secret"]

        # Perform refresh
        client = WebApplicationClient(client_id, refresh_token=refresh_token)
        uri, headers, body = client.prepare_refresh_token_request(token_endpoint)
        refresh_result = requests.post(uri, headers=headers, data=body,
                                       auth=(client_id, client_secret))

        if refresh_result.ok:
            new_token_data = refresh_result.json()
            # Make sure 'expires_at' entry exists in token
            if 'expires_at' not in new_token_data:
                new_token_data['expires_at'] = time.time() + token_data['expires_in']
            # Store token
            with OAuthTokenDatabase() as db:
                token_data.update(new_token_data)
                db.update_token(dev_cred, token_data)

        return token_data

    def oauth_get_token(self, dev_cred, auto_refresh=True):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier
                auto_refresh: refresh expired tokens automatically

            Returns:
                json string containing token and additional information
        """
        if auto_refresh:
            expired = False
            with OAuthTokenDatabase() as db:
                token_data = db.get(dev_cred)
            if "expires_at" not in token_data:
                expired = True
            elif token_data["expires_at"] >= time.time():
                expired = True
            if expired:
                return self.oauth_refresh_token(dev_cred)
        return self.db_get_oauth_token(dev_cred)

    # Admin API
    def admin_update_backend_config(self, config):
        LOG.error("tried to update backend config, but not using a backend. ignoring")

    def admin_get_backend_config(self):
        LOG.error("tried to get backend config, but not using a backend")
        return {}

    def admin_pair(self, uuid=None):
        uuid = uuid or str(uuid4())
        # create dummy identity file for third parties expecting it for pairing checks
        identity = {"uuid": uuid,
                    "access": "OVOSdbF1wJ4jA5lN6x6qmVk_QvJPqBQZTUJQm7fYzkDyY_Y=",
                    "refresh": "OVOS66c5SpAiSpXbpHlq9HNGl1vsw_srX49t5tCv88JkhuE=",
                    "expires_at": time.time() + 9999999999}
        # save identity file
        IdentityManager.save(identity)
        return identity

    def admin_set_device_location(self, uuid, loc):
        """
        loc = {
            "city": {
                "code": "Lawrence",
                "name": "Lawrence",
                "state": {
                    "code": "KS",
                    "name": "Kansas",
                    "country": {
                        "code": "US",
                        "name": "United States"
                    }
                }
            },
            "coordinate": {
                "latitude": 38.971669,
                "longitude": -95.23525
            },
            "timezone": {
                "code": "America/Chicago",
                "name": "Central Standard Time",
                "dstOffset": 3600000,
                "offset": -21600000
            }
        }
        """
        update_mycroft_config({"location": loc})

    def admin_set_device_prefs(self, uuid, prefs):
        """
        prefs = {"time_format": "full",
                "date_format": "DMY",
                "system_unit": "metric",
                "lang": "en-us",
                "wake_word": "hey_mycroft",
                "ww_config": {"phonemes": "HH EY . M AY K R AO F T",
                             "module": "ovos-ww-plugin-pocketsphinx",
                             "threshold": 1e-90},
                "tts_module": "ovos-tts-plugin-mimic",
                "tts_config": {"voice": "ap"}}
        """
        cfg = dict(prefs)
        cfg["listener"] = {}
        cfg["hotwords"] = {}
        cfg["tts"] = {}
        tts = None
        tts_cfg = {}
        ww = None
        ww_cfg = {}
        if "wake_word" in cfg:
            ww = cfg.pop("wake_word")
        if "ww_config" in cfg:
            ww_cfg = cfg.pop("ww_config")
        if "tts_module" in cfg:
            tts = cfg.pop("tts_module")
        if "tts_config" in cfg:
            tts_cfg = cfg.pop("tts_config")
            if not tts:
                tts = tts_cfg.get("module")
        if tts:
            cfg["tts"]["module"] = tts
            cfg["tts"][tts] = tts_cfg
        if ww:
            cfg["listener"]["wake_word"] = ww
            cfg["hotwords"][ww] = ww_cfg
        update_mycroft_config(cfg)

    def admin_set_device_info(self, uuid, info):
        """
        info = {"opt_in": True,
                "name": "my_device",
                "device_location": "kitchen",
                "email": "notifications@me.com",
                "isolated_skills": False,
                "lang": "en-us"}
        """
        update_mycroft_config({"opt_in": info["opt_in"],
                               "email": {"recipient": info.get("email")},
                               "lang": info["lang"]})

    # STT Api
    def load_stt_plugin(self, config=None, lang=None):
        from ovos_plugin_manager.stt import OVOSSTTFactory, get_stt_config

        config = config or get_stt_config(config)
        if lang:
            config["lang"] = lang
        self.stt = OVOSSTTFactory.create(config)

    def stt_get(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum alternate transcriptions

       """
        if self.stt is None:
            self.load_stt_plugin(lang=language)
        from speech_recognition import AudioFile, Recognizer
        with NamedTemporaryFile() as fp:
            fp.write(audio)
            with AudioFile(fp.name) as source:
                audio = Recognizer().record(source)
        tx = self.stt.execute(audio, language)
        if isinstance(tx, str):
            tx = [tx]
        return tx

    # Database API
    def db_list_devices(self):
        _mail_cfg = self.credentials.get("email", {})

        tts_plug = Configuration().get("tts").get("module")
        tts_config = Configuration().get("tts")[tts_plug]

        default_ww = Configuration().get("listener").get("wake_word", "hey_mycroft")
        ww_config = Configuration().get("hotwords")[default_ww]

        device = {
            "uuid": self.uuid,
            "token": "DUMMYTOKEN123",
            "isolated_skills": True,
            "opt_in": Configuration().get("opt_in", False),
            "name": f"Device-{self.uuid}",
            "device_location": "somewhere",
            "email": _mail_cfg.get("recipient") or
                     _mail_cfg.get("smtp", {}).get("username"),
            "time_format": Configuration().get("time_format", "full"),
            "date_format": Configuration().get("date_format", "DMY"),
            "system_unit": Configuration().get("system_unit", "metric"),
            "lang": Configuration().get("lang") or "en-us",
            "location": Configuration().get("location"),
            "default_tts": tts_plug,
            "default_tts_cfg": tts_config,
            "default_ww": default_ww,
            "default_ww_cfg": ww_config
        }
        return [device]

    def db_get_device(self, uuid):
        if uuid != self.uuid:
            return None
        return self.db_list_devices()[0]

    def db_update_device(self, uuid, name=None,
                         device_location=None, opt_in=None,
                         location=None, lang=None, date_format=None,
                         system_unit=None, time_format=None, email=None,
                         isolated_skills=False, ww_id=None, voice_id=None):
        if uuid != self.uuid:
            identity = self.admin_pair(uuid)
        new_config = {

        }
        if opt_in is not None:
            new_config["opt_in"] = opt_in
        if location is not None:
            new_config["location"] = location
        if lang is not None:
            new_config["lang"] = lang
        if time_format is not None:
            new_config["time_format"] = time_format
        if date_format is not None:
            new_config["date_format"] = date_format
        if system_unit is not None:
            new_config["system_unit"] = system_unit
        if email is not None:
            new_config["email"]["recipient"] = email
        if device_location is not None:
            pass  # not tracked locally, reserved for future usage
        if ww_id is not None:
            ww_def = self.db_get_ww_definition(ww_id)
            if ww_def:
                name = ww_def["name"]
                cfg = ww_def["ww_config"]
                if "module" not in cfg:
                    cfg["module"] = ww_def["plugin"]
                new_config["listener"]["wake_word"] = name
                new_config["hotwords"][name] = cfg

        if voice_id is not None:
            ww_def = self.db_get_voice_definition(voice_id)
            if ww_def:
                plugin = ww_def["plugin"]
                cfg = ww_def["tts_config"]
                # TODO - gender -> persona
                if ww_def.get("lang"):
                    cfg["lang"] = ww_def["lang"]
                new_config["tts"]["module"] = plugin
                new_config["tts"][plugin] = cfg
        update_mycroft_config(new_config)

    def db_delete_device(self, uuid):
        # delete identity file/user config/skill settings

        settings_path = f"{get_xdg_config_save_path()}/skills"

        skill_ids = listdir(settings_path)

        # delete skill settings
        for skill_id in skill_ids:
            s = f"{settings_path}/{skill_id}/settings.json"
            if isfile(s):
                remove(s)

        if isfile(IdentityManager.IDENTITY_FILE):
            remove(IdentityManager.IDENTITY_FILE)

        if isfile(USER_CONFIG):
            remove(USER_CONFIG)

    def db_post_device(self, uuid, token, *args, **kwargs):
        return self.db_update_device(uuid, *args, **kwargs)

    def db_list_shared_skill_settings(self):
        return [s.serialize() for s in get_local_settings()]

    def db_get_shared_skill_settings(self, skill_id):
        settings = get_local_settings()
        skill_settings = []
        for s in settings:
            if s.skill_id == skill_id:
                skill_settings.append(s.serialize())
        return skill_settings

    def db_update_shared_skill_settings(self, skill_id,
                                        display_name=None,
                                        settings_json=None,
                                        metadata_json=None):
        settings_path = f"{get_xdg_config_save_path()}/skills/{skill_id}"
        makedirs(settings_path, exist_ok=True)
        if metadata_json:
            s = f"{settings_path}/settingsmeta.json"
            with open(s, "w") as f:
                json.dump(metadata_json, f)
        if settings_json:
            s = f"{settings_path}/settings.json"
            with open(s, "w") as f:
                json.dump(metadata_json, f)
        return SkillSettingsModel(skill_id=skill_id,
                                  skill_settings=settings_path,
                                  meta=metadata_json,
                                  display_name=display_name).serialize()

    def db_delete_shared_skill_settings(self, skill_id):
        settings_path = f"{get_xdg_config_save_path()}/skills/{skill_id}"
        deleted = False
        s = f"{settings_path}/settingsmeta.json"
        if isfile(s):
            remove(s)
            deleted = True
        s = f"{settings_path}/settings.json"
        if isfile(s):
            remove(s)
            deleted = True
        return deleted

    def db_post_shared_skill_settings(self, skill_id,
                                      display_name,
                                      settings_json,
                                      metadata_json):
        return self.db_update_shared_skill_settings(skill_id,
                                                    display_name=display_name,
                                                    settings_json=settings_json,
                                                    metadata_json=metadata_json)

    def db_list_skill_settings(self, uuid):
        return self.db_list_shared_skill_settings()

    def db_get_skill_settings(self, uuid, skill_id):
        return self.db_get_shared_skill_settings(skill_id)

    def db_update_skill_settings(self, uuid, skill_id,
                                 display_name=None,
                                 settings_json=None,
                                 metadata_json=None):
        return self.db_update_shared_skill_settings(skill_id, display_name=display_name,
                                                    settings_json=settings_json, metadata_json=metadata_json)

    def db_delete_skill_settings(self, uuid, skill_id):
        return self.db_delete_shared_skill_settings(skill_id)

    def db_post_skill_settings(self, uuid, skill_id,
                               display_name,
                               settings_json,
                               metadata_json):
        return self.db_post_shared_skill_settings(skill_id, display_name=display_name,
                                                  settings_json=settings_json, metadata_json=metadata_json)

    def db_list_oauth_apps(self):
        return OAuthApplicationDatabase().values()

    def db_get_oauth_app(self, token_id):
        return OAuthApplicationDatabase().get_application(token_id)

    def db_update_oauth_app(self, token_id, client_id=None, client_secret=None,
                            auth_endpoint=None, token_endpoint=None,
                            callback_endpoint=None, scope=None, shell_integration=None):
        with OAuthApplicationDatabase() as db:
            return db.add_token(token_id, client_id, client_secret,
                                auth_endpoint, token_endpoint,
                                callback_endpoint, scope, shell_integration)

    def db_delete_oauth_app(self, token_id):
        with OAuthApplicationDatabase() as db:
            return db.delete_application(token_id)

    def db_post_oauth_app(self, token_id, client_id, client_secret,
                          auth_endpoint, token_endpoint,
                          callback_endpoint, scope, shell_integration=True):
        with OAuthApplicationDatabase() as db:
            return db.add_application(token_id, client_id, client_secret,
                                      auth_endpoint, token_endpoint, callback_endpoint,
                                      scope, shell_integration)

    def db_list_oauth_tokens(self):
        return OAuthTokenDatabase().values()

    def db_get_oauth_token(self, token_id):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        return OAuthTokenDatabase().get_token(token_id)

    def db_update_oauth_token(self, token_id, token_data):
        with OAuthTokenDatabase() as db:
            return db.add_token(token_id, token_data)

    def db_delete_oauth_token(self, token_id):
        with OAuthTokenDatabase() as db:
            return db.delete_token(token_id)

    def db_post_oauth_token(self, token_id, token_data):
        with OAuthTokenDatabase() as db:
            return db.add_token(token_id, token_data)

    def db_list_stt_recordings(self):
        return JsonUtteranceDatabase().values()

    def db_get_stt_recording(self, rec_id):
        return JsonUtteranceDatabase().get_utterance(rec_id).serialize()

    def db_update_stt_recording(self, rec_id, transcription=None, metadata=None):
        # TODO - metadata unused, extend db
        return JsonUtteranceDatabase().update_utterance(rec_id, transcription)

    def db_delete_stt_recording(self, rec_id):
        return JsonUtteranceDatabase().delete_utterance(rec_id)

    def db_post_stt_recording(self, byte_data, transcription, metadata=None):
        # TODO - metadata unused, extend db
        save_path = Configuration().get("listener", {}).get('save_path') or \
                    f"{get_xdg_data_save_path()}/listener/utterances"
        os.makedirs(save_path, exist_ok=True)

        with JsonUtteranceDatabase() as db:
            n = f"{transcription.lower().replace('/', '_').replace(' ', '_')}_{db.total_utterances() + 1}"
            with open(f"{save_path}/{n}.wav", "wb") as f:
                f.write(byte_data)
            return db.add_utterance(transcription, f"{save_path}/{n}.wav", self.uuid)

    def db_list_ww_recordings(self):
        return JsonWakeWordDatabase().values()

    def db_get_ww_recording(self, rec_id):
        return JsonWakeWordDatabase().get_wakeword(rec_id).serialize()

    def db_update_ww_recording(self, rec_id, transcription=None, metadata=None):
        with JsonWakeWordDatabase() as db:
            db.update_wakeword(rec_id, transcription=transcription, meta=metadata)

    def db_delete_ww_recording(self, rec_id):
        with JsonWakeWordDatabase() as db:
            db.delete_wakeword(rec_id)

    def db_post_ww_recording(self, byte_data, transcription, metadata=None):
        listener_config = Configuration().get("listener", {})
        save_path = listener_config.get('save_path', f"{get_xdg_data_save_path()}/listener/wake_words")
        filename = join(save_path, '_'.join(str(metadata[k]) for k in sorted(metadata)) + '.wav')
        os.makedirs(save_path, exist_ok=True)
        metadata = metadata or {}
        with open(save_path, "wb") as f:
            f.write(byte_data)
        with JsonWakeWordDatabase() as db:
            db.add_wakeword(metadata["name"], filename, metadata, self.uuid)

    def db_list_metrics(self):
        return JsonMetricDatabase().values()

    def db_get_metric(self, metric_id):
        return JsonMetricDatabase().get(metric_id)

    def db_update_metric(self, metric_id, metadata):
        m = self.db_get_metric(metric_id)
        m.meta = metadata
        with JsonMetricDatabase() as db:
            db[metric_id] = m

    def db_delete_metric(self, metric_id):
        with JsonMetricDatabase() as db:
            if metric_id in db:
                db.pop(metric_id)
                return True
        return False

    def db_post_metric(self, metric_type, metadata):
        with JsonMetricDatabase() as db:
            m = db.add_metric(metric_type, metadata, self.uuid)
        return m.serialize()

    def db_list_ww_definitions(self):
        ww_defs = []
        for ww_id, ww_cfg in get_wws().items():  # TODO scan=True once implemented
            plugin, name, _ = ww_id.split("_", 3)
            ww_defs.append({
                "ww_id": ww_id,
                "name": ww_cfg.get("display_name") or name,
                "lang": ww_cfg.get("stt_lang") or
                        ww_cfg.get("lang") or
                        Configuration().get("lang", "en-us"),
                "plugin": ww_cfg.get("module") or plugin,
                "ww_config": ww_cfg
            })
        return ww_defs

    def db_get_ww_definition(self, ww_id):
        ww_defs = self.db_list_ww_definitions()
        for ww in ww_defs:
            if ww["ww_id"] == ww_id:
                return ww

    def db_update_ww_definition(self, ww_id, name=None, lang=None, ww_config=None, plugin=None):
        ww_folders = f"{xdg_data_home()}/OPM/ww_configs"
        path = ""
        if not lang:
            for l in listdir(ww_folders):
                if isfile(f"{ww_folders}/{l}/{ww_id}.json"):
                    path = f"{ww_folders}/{l}/{ww_id}.json"
                    break
            else:
                lang = Configuration().get("lang")

        path = path or f"{ww_folders}/{lang}/{ww_id}.json"

        ww_config = ww_config or {}
        if isfile(path):
            with open(path) as f:
                old_cfg = json.load(f)
            if ww_config:
                old_cfg.update(ww_config)
            if plugin:
                old_cfg["module"] = plugin
            if name:
                old_cfg["display_name"] = name
            if lang:
                old_cfg["stt_lang"] = lang
        with open(path, "w") as f:
            json.dump(ww_config, f, indent=4, ensure_ascii=False)

    def db_delete_ww_definition(self, ww_id):
        for lang in listdir(f"{xdg_data_home()}/OPM/ww_configs"):
            if f"_{lang}_" in ww_id:
                path = f"{xdg_data_home()}/OPM/ww_configs/{lang}/{ww_id}.json"
                if isfile(path):
                    remove(path)
                    return True
        return False

    def db_post_ww_definition(self, name, lang, ww_config, plugin):
        ww_id = get_ww_id(plugin_name=plugin, ww_name=name, ww_config=ww_config)
        path = f"{xdg_data_home()}/OPM/ww_configs/{lang}/{ww_id}.json"
        ww_config["stt_lang"] = lang  # tag language in STT step
        with open(path, "w") as f:
            json.dump(ww_config, f, indent=4, ensure_ascii=False)

    def db_list_voice_definitions(self):
        return [{
            "voice_id": voice_id,
            "lang": voice_data["meta"].get("lang"),
            "plugin": voice_data["module"],
            "tts_config": voice_data,
            "offline": voice_data["meta"].get("offline"),
            "gender": voice_data["meta"].get("gender"),
        } for voice_id, voice_data in get_voices(scan=True).items()]

    def db_get_voice_definition(self, voice_id):
        voices = get_voices(scan=True)
        if voice_id in voices:
            voice_data = voices[voice_id]
            return {
                "voice_id": voice_id,
                "lang": voice_data["meta"].get("lang"),
                "plugin": voice_data["module"],
                "tts_config": voice_data,
                "offline": voice_data["meta"].get("offline"),
                "gender": voice_data["meta"].get("gender"),
            }
        return {}

    def db_update_voice_definition(self, voice_id, name=None, lang=None, plugin=None,
                                   tts_config=None, offline=None, gender=None):
        VOICES_FOLDER = f"{xdg_data_home()}/OPM/voice_configs"
        if not lang:
            for l in listdir(VOICES_FOLDER):
                if f"_{l}_" in voice_id:
                    lang = l
                    break

        path = f"{VOICES_FOLDER}/{lang}/{voice_id}.json"
        voicedef = {"meta": {}}
        if isfile(path):
            with open(path) as f:
                voicedef = json.load(f)
        if tts_config:
            voicedef.update(tts_config)
        if name:
            voicedef["meta"]["name"] = name
        if name:
            voicedef["meta"]["offline"] = offline
        if gender:
            voicedef["meta"]["gender"] = gender
        if lang:
            voicedef["lang"] = lang
            voicedef["meta"]["lang"] = lang
        if plugin:
            voicedef["module"] = plugin

        with open(path, "w") as f:
            json.dump(voicedef, f, indent=4, ensure_ascii=False)

    def db_delete_voice_definition(self, voice_id):
        VOICES_FOLDER = f"{xdg_data_home()}/OPM/voice_configs"
        for lang in listdir(VOICES_FOLDER):
            if f"_{lang}_" in voice_id:
                path = f"{VOICES_FOLDER}/{lang}/{voice_id}.json"
                if isfile(path):
                    remove(path)
                    return True
        return False

    def db_post_voice_definition(self, name, lang, plugin,
                                 tts_config, offline, gender=None):
        voice_id = get_voice_id(plugin_name=plugin, lang=lang, tts_config=tts_config)
        path = f"{xdg_data_home()}/OPM/voice_configs/{lang}/{voice_id}.json"
        with open(path, "w") as f:
            tts_config["lang"] = lang
            tts_config["meta"] = {"offline": offline, "gender": gender,
                                  "name": name, "lang": lang}
            json.dump(tts_config, f, indent=4, ensure_ascii=False)


class AbstractPartialBackend(OfflineBackend):
    """ helper class that internally delegates unimplemented methods to offline backend implementation
    backends that only provide microservices and no DeviceApi should subclass from here
    """

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=BackendType.OFFLINE, credentials=None):
        super().__init__(url, version, identity_file, credentials)
        self.backend_type = backend_type


if __name__ == "__main__":
    b = OfflineBackend()
    l = b.ip_geolocation_get("0.0.0.0")
    print(l)

    b.load_stt_plugin({"module": "ovos-stt-plugin-vosk"})
    # a = b.geolocation_get("Fafe")
    # a = b.wolfram_full_results("2+2")
    # a = b.wolfram_spoken("what is the speed of light")
    # a = b.owm_get_weather()

    from speech_recognition import Recognizer, AudioFile

    with AudioFile("/home/user/PycharmProjects/selene_api/test/test.wav") as source:
        audio = Recognizer().record(source)

    flac_data = audio.get_flac_data()
    a = b.stt_get(flac_data)

    # a = b.owm_get_weather()
    # a = b.owm_get_daily()
    # a = b.owm_get_hourly()
    # a = b.owm_get_current()
    print(a)



import json
from os import makedirs

from ovos_backend_client.backends import OfflineBackend, \
    PersonalBackend, BackendType, get_backend_config, API_REGISTRY
from ovos_backend_client.settings import get_local_settings
from ovos_config.config import Configuration, get_xdg_config_save_path
from ovos_utils import timed_lru_cache
from ovos_utils.log import LOG


class BaseApi:
    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None, credentials=None):
        url, version, identity_file, backend_type = get_backend_config(url, version,
                                                                       identity_file, backend_type)
        self.url = url
        self.credentials = credentials or {}
        if backend_type == BackendType.PERSONAL:
            self.backend = PersonalBackend(url, version, identity_file, credentials=credentials)
        else:  # if backend_type == BackendType.OFFLINE:
            self.backend = OfflineBackend(url, version, identity_file, credentials=credentials)
        self.validate_backend_type()

    def validate_backend_type(self):
        pass

    @property
    def backend_type(self):
        return self.backend.backend_type

    @property
    def backend_url(self):
        if not self.backend.url.startswith("http"):
            self.backend.url = f"http://{self.backend.url}"
        return self.backend.url

    @property
    def backend_version(self):
        return self.backend.backend_version

    @property
    def identity(self):
        return self.backend.identity

    @property
    def uuid(self):
        return self.backend.uuid

    @property
    def access_token(self):
        return self.backend.access_token

    @property
    def headers(self):
        return self.backend.headers

    def check_token(self):
        self.backend.check_token()

    def refresh_token(self):
        self.backend.refresh_token()

    def get(self, url=None, *args, **kwargs):
        return self.backend.get(url, *args, **kwargs)

    def post(self, url=None, *args, **kwargs):
        return self.backend.post(url, *args, **kwargs)

    def put(self, url=None, *args, **kwargs):
        return self.backend.put(url, *args, **kwargs)

    def patch(self, url=None, *args, **kwargs):
        return self.backend.patch(url, *args, **kwargs)


class AdminApi(BaseApi):
    def __init__(self, admin_key, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type, credentials={"admin": admin_key})
        self.url = f"{self.backend_url}/{self.backend_version}/admin"

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["admin"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def get_backend_config(self):
        return self.backend.admin_get_backend_config()

    def update_backend_config(self, config):
        return self.backend.admin_update_backend_config(config)

    def pair(self, uuid=None):
        return self.backend.admin_pair(uuid)

    def set_device_location(self, uuid, loc):
        """
        loc = {
            "city": {
                "code": "Lawrence",
                "name": "Lawrence",
                "state": {
                    "code": "KS",
                    "name": "Kansas",
                    "country": {
                        "code": "US",
                        "name": "United States"
                    }
                }
            },
            "coordinate": {
                "latitude": 38.971669,
                "longitude": -95.23525
            },
            "timezone": {
                "code": "America/Chicago",
                "name": "Central Standard Time",
                "dstOffset": 3600000,
                "offset": -21600000
            }
        }
        """
        return self.backend.admin_set_device_location(uuid, loc)

    def set_device_prefs(self, uuid, prefs):
        """
        prefs = {"time_format": "full",
                "date_format": "DMY",
                "system_unit": "metric",
                "lang": "en-us",
                "wake_word": "hey_mycroft",
                "ww_config": {"phonemes": "HH EY . M AY K R AO F T",
                             "module": "ovos-ww-plugin-pocketsphinx",
                             "threshold": 1e-90},
                "tts_module": "ovos-tts-plugin-mimic",
                "tts_config": {"voice": "ap"}}
        """
        self.backend.admin_set_device_prefs(uuid, prefs)

    def set_device_info(self, uuid, info):
        """
        info = {"opt_in": True,
                "name": "my_device",
                "device_location": "kitchen",
                "email": "notifications@me.com",
                "isolated_skills": False,
                "lang": "en-us"}
        """
        self.backend.admin_set_device_info(uuid, info)


class DeviceApi(BaseApi):
    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)
        self.url = f"{self.backend_url}/{self.backend_version}/device"

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["device"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def get(self, url=None, *args, **kwargs):
        """ Retrieve all device information from the web backend """
        return self.backend.device_get()

    def get_settings(self):
        """ Retrieve device settings information from the web backend

        Returns:
            str: JSON string with user configuration information.
        """
        return self.backend.device_get_settings()

    def get_code(self, state=None):
        return self.backend.device_get_code(state)

    def activate(self, state, token,
                 core_version="unknown",
                 platform="unknown",
                 platform_build="unknown",
                 enclosure_version="unknown"):
        return self.backend.device_activate(state, token, core_version,
                                            platform, platform_build, enclosure_version)

    def update_version(self,
                       core_version="unknown",
                       platform="unknown",
                       platform_build="unknown",
                       enclosure_version="unknown"):
        return self.backend.device_update_version(core_version, platform, platform_build, enclosure_version)

    def get_location(self):
        """ Retrieve device location information from the web backend

        Returns:
            str: JSON string with user location.
        """
        return self.backend.device_get_location()

    def upload_skill_metadata(self, settings_meta):
        """Upload skill metadata.

        Args:
            settings_meta (dict): skill info and settings in JSON format
        """
        return self.backend.device_upload_skill_metadata(settings_meta)

    def upload_skills_data(self, data):
        """ Upload skills.json file. This file contains a manifest of installed
        and failed installations for use with the Marketplace.

        Args:
             data: dictionary with skills data from msm
        """
        if not isinstance(data, dict):
            raise ValueError('data must be of type dict')

        _data = dict(data)  # Make sure the input data isn't modified
        # Strip the skills.json down to the bare essentials
        to_send = {'skills': []}
        if 'blacklist' in _data:
            to_send['blacklist'] = _data['blacklist']
        else:
            LOG.warning('skills manifest lacks blacklist entry')
            to_send['blacklist'] = []

        # Make sure skills doesn't contain duplicates (keep only last)
        if 'skills' in _data:
            skills = {s['name']: s for s in _data['skills']}
            to_send['skills'] = [skills[key] for key in skills]
        else:
            LOG.warning('skills manifest lacks skills entry')
            to_send['skills'] = []

        for s in to_send['skills']:
            # Remove optional fields backend objects to
            if 'update' in s:
                s.pop('update')

            # Finalize skill_gid with uuid if needed
            s['skill_gid'] = s.get('skill_gid', '').replace('@|', f'@{self.uuid}|')

        return self.backend.device_upload_skills_data(to_send)

    ## DEPRECATED APIS below, use dedicated classes instead
    def get_oauth_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use OAuthApi class instead")
        return self.backend.device_get_oauth_token(dev_cred)

    def report_metric(self, name, data):
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use MetricsApi class instead")
        return self.backend.device_report_metric(name, data)

    def get_subscription(self):
        """
            Get information about type of subscription this unit is connected
            to.

            Returns: dictionary with subscription information
        """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: there are no subscriptions")
        return self.backend.device_get_subscription()

    @property
    def is_subscriber(self):
        """
            status of subscription. True if device is connected to a paying
            subscriber.
        """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: there are no subscriptions")
        return self.backend.is_subscriber

    def get_subscriber_voice_url(self, voice=None, arch=None):
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: there are no subscriptions")
        return self.backend.device_get_subscriber_voice_url(voice, arch)

    def get_skill_settings_v1(self):
        """ old style deprecated bidirectional skill settings api, still available! """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use SkillSettingsApi class instead")
        return self.backend.device_get_skill_settings_v1()

    def put_skill_settings_v1(self, data):
        """ old style deprecated bidirectional skill settings api, still available! """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use SkillSettingsApi class instead")
        return self.backend.device_put_skill_settings_v1(data)

    # cached for 30 seconds because often 1 call per skill is done in quick succession
    @timed_lru_cache(seconds=30)
    def get_skill_settings(self):
        """Get the remote skill settings for all skills on this device."""
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use SkillSettingsApi class instead")
        return self.backend.device_get_skill_settings()

    def send_email(self, title, body, sender):
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use EmailApi class instead")
        return self.backend.device_send_email(title, body, sender)

    def upload_wake_word_v1(self, audio, params):
        """ upload precise wake word V1 endpoint - DEPRECATED"""
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use DatasetApi class instead")
        return self.backend.device_upload_wake_word_v1(audio, params)

    def upload_wake_word(self, audio, params):
        """ upload precise wake word V2 endpoint """
        ## DEPRECATED - compat only for old devices
        LOG.warning("DEPRECATED: use DatasetApi class instead")
        return self.backend.device_upload_wake_word(audio, params)


class STTApi(BaseApi):
    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)
        self.url = f"{self.backend_url}/{self.backend_version}/stt"

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["stt"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    @property
    def headers(self):
        h = self.backend.headers
        h["Content-Type"] = "audio/x-flac"
        return h

    def stt(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum minutes to transcribe(?)

        Returns:
            dict: JSON structure with transcription results
        """
        return self.backend.stt_get(audio, language, limit)


class GeolocationApi(BaseApi):
    """Web API wrapper for performing geolocation lookups."""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["geolocate"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")
        if self.backend_type == BackendType.OFFLINE:
            self.url = "https://nominatim.openstreetmap.org"
        else:
            self.url = f"{self.backend_url}/{self.backend_version}/geolocation"

    def get_geolocation(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        return self.backend.geolocation_get(location)

    def get_ip_geolocation(self, ip):
        """Call the geolocation endpoint.

        Args:
            ip (str): the ip address to lookup

        Returns:
            str: JSON structure with lookup results
        """
        return self.backend.ip_geolocation_get(ip)

    def get_reverse_geolocation(self, lat, lon):
        """"Call the reverse geolocation endpoint.

        Args:
            lat (float): latitude
            lon (float): longitude

        Returns:
            str: JSON structure with lookup results
        """
        return self.backend.reverse_geolocation_get(lat, lon)


class WolframAlphaApi(BaseApi):

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None, key=None):
        super().__init__(url, version, identity_file, backend_type, credentials={"wolfram": key})

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["wolfram"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")
        if self.backend_type == BackendType.OFFLINE and not self.credentials["wolfram"]:
            raise ValueError("WolframAlpha api key not set!")

        if self.backend_type == BackendType.OFFLINE:
            self.url = "https://api.wolframalpha.com"
        else:
            self.url = f"{self.backend_url}/{self.backend_version}/wolframAlpha"

    # cached to save api calls, wolfram answer wont change often
    @timed_lru_cache(seconds=60 * 30)
    def spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        return self.backend.wolfram_spoken(query, units, lat_lon, optional_params)

    @timed_lru_cache(seconds=60 * 30)
    def simple(self, query, units="metric", lat_lon=None, optional_params=None):
        return self.backend.wolfram_simple(query, units, lat_lon, optional_params)

    @timed_lru_cache(seconds=60 * 30)
    def full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
            https://products.wolframalpha.com/api/documentation/
            Pods of interest
            - Input interpretation - Wolfram's determination of what is being asked about.
            - Name - primary name of
            """
        return self.backend.wolfram_full_results(query, units, lat_lon, optional_params)


class OpenWeatherMapApi(BaseApi):
    """Use Open Weather Map's One Call API to retrieve weather information"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None, key=None):
        super().__init__(url, version, identity_file, backend_type, credentials={"owm": key})

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["owm"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")
        if self.backend_type == BackendType.OFFLINE and not self.backend.credentials["owm"]:
            raise ValueError("OWM api key not set!")
        if self.backend_type == BackendType.OFFLINE:
            self.url = "https://api.openweathermap.org/data/2.5"
        else:
            self.url = f"{self.backend_url}/{self.backend_version}/owm"

    def owm_language(self, lang: str):
        """
        OWM supports 31 languages, see https://openweathermap.org/current#multi

        Convert Mycroft's language code to OpenWeatherMap's, if missing use english.

        Args:
            lang: The Mycroft language code.
        """
        return self.backend.owm_language(lang)

    # cached to save api calls, owm only updates data every 15mins or so
    @timed_lru_cache(seconds=60 * 10)
    def get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        return self.backend.owm_get_weather(lat_lon, lang, units)

    @timed_lru_cache(seconds=60 * 10)
    def get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        return self.backend.owm_get_current(lat_lon, lang, units)

    @timed_lru_cache(seconds=60 * 10)
    def get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        return self.backend.owm_get_hourly(lat_lon, lang, units)

    @timed_lru_cache(seconds=60 * 10)
    def get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        return self.backend.owm_get_daily(lat_lon, lang, units)


class EmailApi(BaseApi):
    """Web API wrapper for sending email"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["email"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")
        if self.backend_type == BackendType.OFFLINE:
            self.url = self.credentials["smtp"]["host"]
        else:
            self.url = self.backend_url

    def send_email(self, title, body, sender):
        return self.backend.email_send(title, body, sender)


class SkillSettingsApi(BaseApi):
    """Web API wrapper for skill settings"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["skill_settings"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def upload_skill_settings(self):
        """ upload skill settings from XDG path"""
        return self.backend.skill_settings_upload(get_local_settings())

    def download_skill_settings(self):
        """ write downloaded settings to XDG path"""
        settings = self.backend.skill_settings_download()
        for s in settings:  # list of SkillSettingsModel or dicts
            settings_path = f"{get_xdg_config_save_path()}/skills/{s.skill_id}"
            makedirs(settings_path, exist_ok=True)
            with open(f"{settings_path}/settingsmeta.json", "w") as f:
                json.dump(s.meta, f, indent=4, ensure_ascii=False)
            with open(f"{settings_path}/settings.json", "w") as f:
                json.dump(s.skill_settings, f, indent=4, ensure_ascii=False)
        return settings


class DatasetApi(BaseApi):
    """Web API wrapper for dataset collection"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["dataset"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def upload_wake_word(self, audio, params, upload_url=None):
        return self.backend.dataset_upload_wake_word(audio, params, upload_url)

    def upload_stt_recording(self, audio, params, upload_url=None):
        return self.backend.dataset_upload_stt_recording(audio, params, upload_url)


class MetricsApi(BaseApi):
    """Web API wrapper for netrics collection"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["metrics"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def report_metric(self, name, data):
        return self.backend.metrics_upload(name, data)


class OAuthApi(BaseApi):
    """Web API wrapper for oauth api"""

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type)

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["oauth"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")

    def refresh_oauth_token(self, dev_cred):
        """
            Refresh Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        return self.backend.oauth_refresh_token(dev_cred)

    def get_oauth_token(self, dev_cred, auto_refresh=True):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier
                auto_refresh: refresh expired tokens automatically

            Returns:
                json string containing token and additional information
        """
        return self.backend.oauth_get_token(dev_cred, auto_refresh=auto_refresh)


class DatabaseApi(BaseApi):
    """Web API wrapper for oauth api"""

    def __init__(self, admin_key=None, url=None, version="v1", identity_file=None, backend_type=None):
        super().__init__(url, version, identity_file, backend_type, credentials={"admin": admin_key})
        self.url = f"{self.backend_url}/{self.backend_version}/admin"

    def validate_backend_type(self):
        if not API_REGISTRY[self.backend_type]["database"]:
            raise ValueError(f"{self.__class__.__name__} not available for {self.backend_type}")
        if self.backend_type in [BackendType.PERSONAL] and not self.credentials.get("admin"):
            raise ValueError(f"Admin key not set, can not access remote database")

    def list_devices(self):
        return self.backend.db_list_devices()

    def get_device(self, uuid):
        return self.backend.db_get_device(uuid)

    def update_device(self, uuid, name=None,
                      device_location=None, opt_in=False,
                      location=None, lang=None, date_format=None,
                      system_unit=None, time_format=None, email=None,
                      isolated_skills=False, ww_id=None, voice_id=None):
        return self.backend.db_update_device(uuid, name, device_location, opt_in,
                                             location, lang, date_format, system_unit, time_format,
                                             email, isolated_skills, ww_id, voice_id)

    def delete_device(self, uuid):
        return self.backend.db_delete_device(uuid)

    def add_device(self, uuid, token, name=None,
                   device_location="somewhere",
                   opt_in=Configuration().get("opt_in", False),
                   location=Configuration().get("location"),
                   lang=Configuration().get("lang"),
                   date_format=Configuration().get("date_format", "DMY"),
                   system_unit=Configuration().get("system_unit", "metric"),
                   time_format=Configuration().get("date_format", "full"),
                   email=None,
                   isolated_skills=False,
                   ww_id=None,
                   voice_id=None):
        return self.backend.db_post_device(uuid, token, name, device_location, opt_in,
                                           location, lang, date_format, system_unit, time_format,
                                           email, isolated_skills, ww_id, voice_id)

    def list_shared_skill_settings(self):
        return self.backend.db_list_shared_skill_settings()

    def get_shared_skill_settings(self, skill_id):
        return self.backend.db_get_shared_skill_settings(skill_id)

    def update_shared_skill_settings(self, skill_id,
                                     display_name=None,
                                     settings_json=None,
                                     metadata_json=None):
        return self.backend.db_update_shared_skill_settings(skill_id, display_name,
                                                            settings_json, metadata_json)

    def delete_shared_skill_settings(self, skill_id):
        return self.backend.db_delete_shared_skill_settings(skill_id)

    def add_shared_skill_settings(self, skill_id,
                                  display_name,
                                  settings_json,
                                  metadata_json):
        return self.backend.db_post_shared_skill_settings(skill_id, display_name,
                                                          settings_json, metadata_json)

    def list_skill_settings(self, uuid):
        return self.backend.db_list_skill_settings(uuid)

    def get_skill_settings(self, uuid, skill_id):
        return self.backend.db_get_skill_settings(uuid, skill_id)

    def update_skill_settings(self, uuid, skill_id,
                              display_name=None,
                              settings_json=None,
                              metadata_json=None):
        return self.backend.db_update_skill_settings(uuid, skill_id, display_name,
                                                     settings_json, metadata_json)

    def delete_skill_settings(self, uuid, skill_id):
        return self.backend.db_delete_skill_settings(uuid, skill_id)

    def add_skill_settings(self, uuid, skill_id,
                           display_name,
                           settings_json,
                           metadata_json):
        return self.backend.db_post_skill_settings(uuid, skill_id, display_name,
                                                   settings_json, metadata_json)

    def list_oauth_apps(self):
        return self.backend.db_list_oauth_apps()

    def get_oauth_app(self, token_id):
        return self.backend.db_get_oauth_app(token_id)

    def update_oauth_app(self, token_id, client_id=None, client_secret=None,
                         auth_endpoint=None, token_endpoint=None,
                         callback_endpoint=None, scope=None, shell_integration=None):
        return self.backend.db_update_oauth_app(token_id, client_id, client_secret, auth_endpoint, token_endpoint,
                                                callback_endpoint, scope, shell_integration)

    def delete_oauth_app(self, token_id):
        return self.backend.db_delete_oauth_app(token_id)

    def add_oauth_app(self, token_id, client_id, client_secret,
                      auth_endpoint, token_endpoint,
                      callback_endpoint, scope, shell_integration=True):
        return self.backend.db_post_oauth_app(token_id, client_id, client_secret, auth_endpoint, token_endpoint,
                                              callback_endpoint, scope, shell_integration)

    def list_oauth_tokens(self):
        return self.backend.db_list_oauth_tokens()

    def get_oauth_token(self, token_id):
        return self.backend.db_get_oauth_token(token_id)

    def update_oauth_token(self, token_id, token_data):
        return self.backend.db_update_oauth_token(token_id, token_data)

    def delete_oauth_token(self, token_id):
        return self.backend.db_delete_oauth_token(token_id)

    def add_oauth_token(self, token_id, token_data):
        return self.backend.db_post_oauth_token(token_id, token_data)

    def list_stt_recordings(self):
        return self.backend.db_list_stt_recordings()

    def get_stt_recording(self, rec_id):
        return self.backend.db_get_stt_recording(rec_id)

    def update_stt_recording(self, rec_id, transcription=None, metadata=None):
        return self.backend.db_update_stt_recording(rec_id, transcription, metadata)

    def delete_stt_recording(self, rec_id):
        return self.backend.db_delete_stt_recording(rec_id)

    def add_stt_recording(self, byte_data, transcription, metadata=None):
        return self.backend.db_post_stt_recording(byte_data, transcription, metadata)

    def list_ww_recordings(self):
        return self.backend.db_list_ww_recordings()

    def get_ww_recording(self, rec_id):
        return self.backend.db_get_ww_recording(rec_id)

    def update_ww_recording(self, rec_id, transcription=None, metadata=None):
        return self.backend.db_update_ww_recording(rec_id, transcription, metadata)

    def delete_ww_recording(self, rec_id):
        return self.backend.db_delete_ww_recording(rec_id)

    def add_ww_recording(self, byte_data, transcription, metadata=None):
        return self.backend.db_post_ww_recording(byte_data, transcription, metadata)

    def list_metrics(self):
        return self.backend.db_list_metrics()

    def get_metric(self, metric_id):
        return self.backend.db_get_metric(metric_id)

    def update_metric(self, metric_id, metadata):
        return self.backend.db_update_metric(metric_id, metadata)

    def delete_metric(self, metric_id):
        return self.backend.db_delete_metric(metric_id)

    def add_metric(self, metric_type, metadata):
        return self.backend.db_post_metric(metric_type, metadata)

    def list_ww_definitions(self):
        return self.backend.db_list_ww_definitions()

    def get_ww_definition(self, ww_id):
        return self.backend.db_get_ww_definition(ww_id)

    def update_ww_definition(self, ww_id, name, lang, ww_config, plugin):
        return self.backend.db_update_ww_definition(ww_id, name, lang, ww_config, plugin)

    def delete_ww_definition(self, ww_id):
        return self.backend.db_delete_ww_definition(ww_id)

    def add_ww_definition(self, name, lang, ww_config, plugin):
        return self.backend.db_post_ww_definition(name, lang, ww_config, plugin)

    def list_voice_definitions(self):
        return self.backend.db_list_voice_definitions()

    def get_voice_definition(self, voice_id):
        return self.backend.db_get_voice_definition(voice_id)

    def update_voice_definition(self, voice_id, name=None, lang=None, plugin=None,
                                tts_config=None, offline=None, gender=None):
        return self.backend.db_update_voice_definition(voice_id, name, lang, plugin, tts_config, offline, gender)

    def delete_voice_definition(self, voice_id):
        return self.backend.db_delete_voice_definition(voice_id)

    def add_voice_definition(self, name, lang, plugin,
                             tts_config, offline, gender=None):
        return self.backend.db_post_voice_definition(name, lang, plugin, tts_config, offline, gender)


if __name__ == "__main__":
    # d = DeviceApi(FAKE_BACKEND_URL)

    # TODO turn these into unittests
    # voice_id = load_identity()
    # paired = is_paired()
    geo = GeolocationApi(backend_type=BackendType.OFFLINE)
    data = geo.get_geolocation("Missouri Kansas")
    print(data)
    exit(6)
    wolf = WolframAlphaApi(backend_type=BackendType.OFFLINE)
    # data = wolf.spoken("what is the speed of light")
    # print(data)
    # data = wolf.full_results("2+2")
    # print(data)

    owm = OpenWeatherMapApi(backend_type=BackendType.OFFLINE)
    data = owm.get_current()
    print(data)
    data = owm.get_weather()
    print(data)
